/*
 *	SCCS: @(#)tet_spawn.c	1.16 (97/07/21)
 *
 *	UniSoft Ltd., London, England
 *
 * (C) Copyright 1996 X/Open Company Limited
 *
 * All rights reserved.  No part of this source code may be reproduced,
 * stored in a retrieval system, or transmitted, in any form or by any
 * means, electronic, mechanical, photocopying, recording or otherwise,
 * except as stated in the end-user licence agreement, without the prior
 * permission of the copyright owners.
 * A copy of the end-user licence agreement is contained in the file
 * Licence which accompanies this distribution.
 * 
 * X/Open and the 'X' symbol are trademarks of X/Open Company Limited in
 * the UK and other countries.
 */

#ifndef lint
static char sccsid[] = "@(#)tet_spawn.c	1.16 (97/07/21) TETware release 3.2";
#endif

/************************************************************************

SCCS:   	@(#)tet_spawn.c	1.16 97/07/21 TETware release 3.2
NAME:		'C' API spawn process function
PRODUCT:	TETware
AUTHOR:		Geoff Clare, UniSoft Ltd.
DATE CREATED:	July 1996
SYNOPSIS:

	pid_t	tet_spawn(char *file, char *argv[], char *envp[]);
	int	tet_wait(pid_t pid, int *statp);
	int	tet_kill(pid_t pid, int sig);

DESCRIPTION:

	Tet_spawn() executes the specified file in a child process.
	The argv[] is passed to the tet_main() routine in the new
	process image.  The child calls tet_setcontext() to distinguish
	it's results file output from the parent's.  A call to
	tet_setblock() is made before returning control in the parent,
	to separate output made before and during execution of the child
	process.  Signals which are being caught in the parent are set
	to SIG_DFL in the child so unexpected signals will come
	through to the wait status.  If the exec() of the specified
	file fails, the child exits with code 127.

	Tet_wait() is used to obtain the wait status of a process
	previously spawned with tet_spawn().  A call to tet_setblock()
	is made, to separate output made during and after execution of
	the child.

	Tet_kill() sends the specified signal to the specified process.

MODIFICATIONS:
	Geoff Clare, UniSoft Ltd., Sept 1996
	Changes for TETWare-Lite.

	Andrew Dingwall, UniSoft Ltd., April 1997
	fixed problem with tet_spawn on Windows NT;
	improved thread-safety in mt environment

	Andrew Dingwall, UniSoft Ltd., May 1997
	port to Windows 95


************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <sys/types.h>
#ifdef _WIN32	/* -START-WIN32-CUT- */
#  include <process.h>
#else		/* -END-WIN32-CUT- */
#  include <unistd.h>
#  include <sys/wait.h>
#  include <signal.h>
#endif		/* -WIN32-CUT-LINE- */
#include "dtmac.h"
#include "dtthr.h"
#include "error.h"
#include "tet_api.h"
#include "apilib.h"

#define	KILLWAIT	10

#ifdef NEEDsrcFile
static char srcFile[] = __FILE__;
#endif

#ifndef _WIN32	/* -WIN32-CUT-LINE- */
static pid_t	childpid;
#endif		/* -WIN32-CUT-LINE- */


#ifndef _WIN32	/* -WIN32-CUT-LINE- */

/* ARGSUSED */
static void
sig_term(sig)
int sig;
{
	/* clean up on receipt of SIGTERM, but arrange for wait
	   status still to show termination by SIGTERM */

	struct sigaction sa;

	if (childpid > 0)
		(void) tet_killw(childpid, KILLWAIT);

	sa.sa_handler = SIG_DFL;
	sa.sa_flags = 0; 
	(void) sigemptyset(&sa.sa_mask); 
	(void) sigaction(SIGTERM, &sa, (struct sigaction *)NULL);
	(void) kill(getpid(), SIGTERM);
}

#endif		/* -WIN32-CUT-LINE- */


pid_t
tet_spawn(file, argv, envp)
char *file;
char *argv[];
char *envp[];
{
	pid_t pid;

#ifdef _WIN32	/* -START-WIN32-CUT- */
	char **newargv = (char **) 0, **newenvp = (char **) 0;
	int rc;
#else		/* -END-WIN32-CUT- */
	int i;
	struct sigaction new_sa; 
#  ifdef TET_THREADS
	int err;
	sigset_t oldset;
	extern thread_t tet_start_tid;
	extern void tet_mtx_lock();
	extern void tet_mtx_unlock();
	extern void tet_thrtab_reset();
#  endif /* TET_THREADS */
	extern int tet_mypid;
#endif		/* -WIN32-CUT-LINE- */

	if (!file || !*file || !argv || !envp) {
		tet_errno = TET_ER_INVAL;
		return(-1);
	}

	(void) fflush(stdout);
	(void) fflush(stderr);

#ifdef _WIN32	/* -START-WIN32-CUT- */

	API_LOCK;
	rc = tet_exec_prep(argv, envp, &newargv, &newenvp);
	API_UNLOCK;

	if (rc < 0)
		pid = -1;
	else {
		ASSERT(newargv != (char **) 0);
		ASSERT(newenvp != (char **) 0);
		(void) _flushall();
		if ((pid = (pid_t) tet_spawnvpe(file, newargv, newenvp)) == -1)
			switch (errno) {
			case E2BIG:
				tet_errno = TET_ER_2BIG;
				break;
			case EINVAL:
				tet_errno = TET_ER_INVAL;
				break;
			case ENOENT:
				tet_errno = TET_ER_NOENT;
				break;
			case ENOEXEC:
			case ENOMEM:
			default:
				error(errno, "spawn failed in tet_spawn():",
					file);
				tet_errno = TET_ER_ERR;
				break;
			}
	}

	tet_exec_cleanup(envp, newargv, newenvp);

#else /* _WIN32 */	/* -END-WIN32-CUT- */

#  ifdef TET_THREADS
	/*
	 * Must obtain all mutexes and locks before fork1(), to ensure
	 * they are not held by threads which will not exist in the
	 * child.  Note that by obtaining tet_top_mtx without going
	 * through API_LOCK we are assuming that this function is never
	 * called from other API functions.
	 */
	err = thr_sigsetmask(SIG_BLOCK, &tet_blockable_sigs, &oldset);
	if (err != 0)
	{
		/* not fatal, as we haven't forked yet */
		error(err, "thr_sigsetmask() failed in tet_spawn()", (char *)0);
		tet_errno = TET_ER_ERR;
		return -1;
	}
	tet_mtx_lock();

	pid = fork1();
#  else /* TET_THREADS */
	pid = fork();
#  endif /* TET_THREADS */

	childpid = pid;
	switch (pid)
	{
	case -1:
#  ifdef TET_THREADS
		tet_mtx_unlock();
		(void) thr_sigsetmask(SIG_SETMASK, &oldset, (sigset_t *)0);
#  endif /* TET_THREADS */
		tet_errno = TET_ER_FORK;
		return -1;

	case 0:
		/* child process */

		tet_child = 0;
		tet_mypid = getpid();
#  ifdef TET_THREADS
		tet_start_tid = thr_self();
		tet_thrtab_reset(); /* empty the thread table */
#  endif /* TET_THREADS */

		/* set signals which were caught in parent to default */

		/* NSIG is not provided by POSIX.1:  it must be
		   defined via an extra feature-test macro, or
		   on the compiler command line */
		for (i = 1; i < NSIG; i++)
		{
			if (sigaction(i, (struct sigaction *)NULL,
						    &new_sa) != -1 &&
				new_sa.sa_handler != SIG_DFL &&
				new_sa.sa_handler != SIG_IGN)
			{
				new_sa.sa_handler = SIG_DFL;
				(void) sigaction(i, &new_sa,
					(struct sigaction *)NULL);
			}
		}

#  ifndef TET_LITE /* -START-LITE-CUT- */
		/* disconnect from all connected servers */
		tet_disconnect();
#  endif /* -END-LITE-CUT- */

#  ifdef TET_THREADS
		tet_mtx_unlock();
		(void) thr_sigsetmask(SIG_SETMASK, &oldset, (sigset_t *)0);
#  endif /* TET_THREADS */

		/* change context to distinguish output from parent's */
		tet_setcontext();
#  ifdef TET_THREADS
		/* threads version of tet_setcontext() does not reset the
		   block number(s), but as we called fork1() we know there
		   is only one thread in the child, so it's OK to reset it. */ 
		tet_next_block = 0;
		tet_setblock();
#  endif /* TET_THREADS */

		/* execute specified file */
		errno = 0;
		(void) tet_exec(file, argv, envp);
		if (errno == ENOMEM)
		{
			/* This message is to distinguish malloc()
			   failure from exec() failure */
			error(errno, "tet_exec() failed:", file);
			(void) fflush(stderr);
		}
		_exit(127);
	}

	/* parent process */

	/* if SIGTERM is set to default (e.g. if this tet_spawn() was
	   called from a child), catch it so we can propagate SIGTERM
	   to the child */
	if (sigaction(SIGTERM, (struct sigaction *)NULL, &new_sa) != -1 &&
		new_sa.sa_handler == SIG_DFL)
	{
		new_sa.sa_handler = sig_term;
		(void) sigaction(SIGTERM, &new_sa, (struct sigaction *)NULL);
	}

#  ifdef TET_THREADS
	tet_mtx_unlock();
	(void) thr_sigsetmask(SIG_SETMASK, &oldset, (sigset_t *)0);
#  endif /* TET_THREADS */

#endif /* _WIN32 */	/* -WIN32-CUT-LINE- */

	tet_setblock();

	return pid;
}

int
tet_wait(pid, statp)
pid_t pid;
int *statp;
{
	pid_t rtpid;
	int err;

#ifndef _WIN32	/* -WIN32-CUT-LINE- */
	struct sigaction new_sa; 
#endif		/* -WIN32-CUT-LINE- */

	/* must wait for a specific process: no zero or -ve pid values
	   that waitpid() would treat specially */
#ifdef _WIN32	/* -START-WIN32-CUT- */
	if (pid == -1 || pid == -2)
#else		/* -END-WIN32-CUT- */
	if (pid <= 0)
#endif		/* -WIN32-CUT-LINE- */
	{
		tet_errno = TET_ER_INVAL;
		return -1;
	}

#ifdef _WIN32	/* -START-WIN32-CUT- */
	rtpid = _cwait(statp, pid, 0);
#else		/* -END-WIN32-CUT- */
	rtpid = waitpid(pid, statp, 0);
#endif		/* -WIN32-CUT-LINE- */
	err = errno;

	if (rtpid == -1)
	{
		switch (err)
		{
		case ECHILD:
			tet_errno = TET_ER_PID;
			break;
		case EINVAL:
			tet_errno = TET_ER_INVAL;
			break;
#ifndef _WIN32	/* -WIN32-CUT-LINE- */
		case EINTR:
			tet_errno = TET_ER_WAIT;
			break;
#endif		/* -WIN32-CUT-LINE- */
		default:
			/* since error() is a macro, we can't just have the
			   _cwait()/waitpid() part conditional on _WIN32
			   (at least on MSVC++) we must repeat the lot!
			*/
#ifdef _WIN32	/* -START-WIN32-CUT- */
			error(err,
				"tet_wait() got unexpected errno value from",
				"_cwait()");
#else		/* -END-WIN32-CUT- */
			error(err,
				"tet_wait() got unexpected errno value from",
				"waitpid()");
#endif		/* -WIN32-CUT-LINE- */
			tet_errno = TET_ER_ERR;
		}
	}
	else
	{
		/* only do these things if the wait succeeds */

		API_LOCK;

		tet_setblock();

#ifndef _WIN32	/* -WIN32-CUT-LINE- */
		/* undo SIGTERM handling */
		if (sigaction(SIGTERM, (struct sigaction *)NULL, &new_sa) != -1 &&
			new_sa.sa_handler == sig_term)
		{
			new_sa.sa_handler = SIG_DFL;
			(void) sigaction(SIGTERM, &new_sa,
				(struct sigaction *)NULL);
		}
#endif		/* -WIN32-CUT-LINE- */

		API_UNLOCK;
	}

	errno = err;
	return (rtpid == -1 ? -1 : 0);
}

int
tet_kill(pid, sig)
pid_t pid;
int sig;
{
	int rtval;

	/* must specify a single process: no zero or -ve pid values
	   that kill() would treat specially */
#ifdef _WIN32	/* -START-WIN32-CUT- */
	if (pid == -1 || pid == -2)
#else		/* -END-WIN32-CUT- */
	if (pid <= 0)
#endif		/* -WIN32-CUT-LINE- */
	{
		tet_errno = TET_ER_INVAL;
		return -1;
	}

	rtval = KILL(pid, sig);
	if (rtval == -1)
	{
		switch (errno)
		{
		case EINVAL:
			tet_errno = TET_ER_INVAL;
			break;
		case ESRCH:
			tet_errno = TET_ER_PID;
			break;
		case EPERM:
			tet_errno = TET_ER_PERM;
			break;
		default:
			error(errno,
			    "tet_kill() got unexpected errno value from KILL()",
			    (char *)0);
			tet_errno = TET_ER_ERR;
		}
	}

	return rtval;
}

